
import logging
from typing import Any, Optional

from brownie import chain
from y import convert
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

@yLazyLogger(logger)
def is_froyo(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return chain.id == Network.Fantom and address == '0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3'

@yLazyLogger(logger)
def get_price(token: Any, block: Optional[Block] = None) -> UsdPrice:
    pool = '0x83E5f18Da720119fF363cF63417628eB0e9fd523'
    return UsdPrice(raw_call(pool, "get_virtual_price()", block=block, output='int') / 1e18)




from typing import Optional

from brownie.convert.datatypes import EthAddress
from y.datatypes import Block, UsdPrice
from y.prices import magic

MAPPING = {
    "0x30D9410ED1D5DA1F6C8391af5338C93ab8d4035C": "0x6c3F90f043a72FA612cbac8115EE7e52BDe6E490", # cvx3crv -> 3crv
    "0xbE0F6478E0E4894CFb14f32855603A083A57c7dA": "0xd632f22692FaC7611d2AA1C0D552930D43CAEd3B", # cvxFRAX3CRV-f -> FRAX3CRV
    "0xabB54222c2b77158CC975a2b715a3d703c256F05": "0x5a6A4D54456819380173272A5E8E9B9904BdF41B", # cvxMIM-3LP3CRV-f -> crvMIM
    "0xCA3D9F45FfA69ED454E66539298709cb2dB8cA61": "0x43b4FdFD4Ff969587185cDB6f0BD875c5Fc83f8c", # cvxalUSD3CRV-f -> crvalusd
}

def is_convex_lp(token_address: EthAddress) -> bool:
    return token_address in MAPPING

def get_price(token_address: EthAddress, block: Optional[Block] = None) -> UsdPrice:
    return magic.get_price(MAPPING[token_address],block)



# async: done

import logging

from brownie import chain
from y.constants import wbtc, weth
from y.networks import Network
from y.prices.utils.buckets import check_bucket_async
from y.utils.logging import yLazyLogger
from y.utils.raw_calls import _symbol_async, raw_call_async

logger = logging.getLogger(__name__)

# This module is far from perfect, but provides an acceptable way to validate some of the prices returned by `get_price`

ACCEPTABLE_HIGH_PRICES = {
    Network.Mainnet: [
        # eth and eth-like
        "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", # eth
        weth.address,
        "0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0", # wsteth
        "0xA3D87FffcE63B53E0d54fAa1cc983B7eB0b74A9c", # ecrv
        "0xaA17A236F2bAdc98DDc0Cf999AbB47D47Fc0A6Cf", # ankrcrv
        "0x53a901d48795C58f485cBB38df08FA96a24669D5", # rcrv
        "0xC4C319E2D4d66CcA4464C0c2B32c9Bd23ebe784e", # alETH+ETH-f
        "0x06325440D014e39736583c165C2963BA99fAf14E", # stecrv
        "0x5e74C9036fb86BD7eCdcb084a0673EFc32eA31cb", # sETH
        "0xae78736Cd615f374D3085123A210448E74Fc6393", # rETH
        "0xE95A203B1a91a908F9B9CE46459d101078c2c3cb", # aETHc
        "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84", # stETH
        # btc and btc-like
        wbtc.address,
        "0xEB4C2781e4ebA804CE9a9803C67d0893436bB27D", # renbtc
        "0xfE18be6b3Bd88A2D2A7f928d00292E7a9963CfC6", # sbtc
        "0x49849C98ae39Fff122806C06791Fa73784FB3675", # crvrenwbtc
        "0x075b1bb99792c9E1041bA13afEf80C91a1e70fB3", # crvrenwsbtc
        "0xb19059ebb43466C323583928285a49f558E572Fd", # hcrv
        "0x64eda51d3Ad40D56b9dFc5554E06F94e1Dd786Fd", # tbtc/sbtcCrv
        "0xDE5331AC4B3630f94853Ff322B66407e0D6331E8", # pBTC/sbtcCRV
        "0x410e3E86ef427e30B9235497143881f717d93c2A", # bBTC/sbtcCRV
        "0x2fE94ea3d5d4a175184081439753DE15AeF9d614", # oBTC/sbtcCRV
        "0x0327112423F3A68efdF1fcF402F6c5CB9f7C33fd", # btc++
        "0x3212b29E33587A00FB1C83346f5dBFA69A458923", # imbtc
        "0x5228a22e72ccC52d415EcFd199F99D0665E7733b", # pbtc
        "0xFbdCA68601f835b27790D98bbb8eC7f05FDEaA9B", # ibbtc/sbtcCRV-f
        "0x0316EB71485b0Ab14103307bf65a021042c6d380", # hbtc
        "0x9BE89D2a4cd102D8Fecc6BF9dA793be995C22541", # bbtc
        # gold tokens
        "0x45804880De22913dAFE09f4980848ECE6EcbAf78", # paxg
        "0x4922a015c4407F87432B179bb209e125432E4a2A", # xaut
        # other
        "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e", # yfi
        "0x1cEB5cB57C4D4E2b2433641b95Dd330A33185A44", # kp3r
        "0xD5525D397898e5502075Ea5E830d8914f6F0affe", # meme
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2", # mkr
        "0x3A283D9c08E8b55966afb64C515f5143cf907611", # crvCVXETH
        "0x23B608675a2B2fB1890d3ABBd85c5775c51691d5", # socks
        "0xcA3d75aC011BF5aD07a98d02f18225F9bD9A6BDF", # crvtricrypto
        "0xc4AD29ba4B3c580e6D59105FFf484999997675Ff", # crv3crypto
        "0xd075e95423C5c4BA1E122CaE0f4CdFA19b82881b", # wPE
        "0xe9F84dE264E91529aF07Fa2C746e934397810334", # sak3
        "0xa1d0E215a23d7030842FC67cE582a6aFa3CCaB83", # yfii
        "0x0fe629d1E84E171f8fF0C1Ded2Cc2221Caa48a3f", # mask
        "0x5F0E628B693018f639D10e4A4F59BD4d8B2B6B44", # white
        # nfts
        "0x641927E970222B10b2E8CDBC96b1B4F427316f16", # meeb
        "0x9cea2eD9e47059260C97d697f82b8A14EfA61EA5", # punk
        "0x269616D549D7e8Eaa82DFb17028d0B212D11232A", # punk (different)
        "0x69BbE2FA02b4D90A944fF328663667DC32786385", # punk-basic
        "0xD70240Dd62F4ea9a6A2416e0073D72139489d2AA", # glyph
        "0x114f1388fAB456c4bA31B1850b244Eedcd024136", # coolcats
        "0xEA47B64e1BFCCb773A0420247C0aa0a3C1D2E5C5", # bayc
    ],
    Network.BinanceSmartChain: [
        # eth and eth-like
        weth.address,
        "0x8d0e18c97e5dd8ee2b539ae8cd3a3654df5d79e5", # bweth
        "0xbfF4a34A4644a113E8200D7F1D79b3555f723AfE", # ibeth
        # btc and btc-like
        wbtc.address,
        "0x7130d2A12B9BCbFAe4f2634d864A1Ee1Ce3Ead9c", # btcb
        "0x08FC9Ba2cAc74742177e0afC3dC8Aed6961c24e7", # ibbtcb
    ],
    Network.Polygon: [
        # eth and eth-like
        weth.address,
        # btc and btc-like
        wbtc.address,
    ],
    Network.Fantom: [
        # eth and eth-like
        weth.address,
        # btc and btc-like
        wbtc.address,
        # other
        "0x29b0Da86e484E1C0029B56e817912d778aC0EC69", # yfi
    ],
    Network.Avalanche: [
        # eth and eth-like
        weth.address,
        # btc and btc-like
        wbtc.address,
        "0xDBf31dF14B66535aF65AaC99C32e9eA844e14501", # renbtc

        "0xd6070ae98b8069de6B494332d1A1a81B6179D960", # bifi
    ],
    Network.Arbitrum: [
        # eth and eth-like
        weth.address,
        # btc and btc-like
        wbtc.address,
    ]
}.get(chain.id, [])

async def _sense_check(
    token_address: str, 
    price: float
    ):

    # if price is in "normal" range, exit sense check
    if price < 1000:
        return None
    
    # if we've already validated that the token should have a high price, exit sense check
    if token_address in ACCEPTABLE_HIGH_PRICES:
        return None

    # for some token types, its normal to have a crazy high nominal price
    # we can skip the sense check for those
    if await _exit_sense_check(token_address):
        return None

    print('this prints')
    # proceed with sense check
    price_readable = round(price, 4)
    symbol = await _symbol_async(token_address)
    network = Network.name(chain.id)
    logger.warning(f'unusually high price (${price_readable}) returned for {symbol} {token_address} on {network}. This does not necessarily mean that the price is wrong, but you may want to validate the price for yourself before proceeding.')


@yLazyLogger(logger)
async def _exit_sense_check(token_address: str) -> bool:
    '''
    For some token types, its normal to have a crazy high nominal price.
    We can skip the sense check for those.
    We can also skip wrapped versions of tokens in `ACCEPTABLE_HIGH_PRICES`.
    '''

    bucket = await check_bucket_async(token_address)

    if bucket == 'uni or uni-like lp':
        return True
    
    elif bucket == 'balancer pool':
        return True
    
    # for wrapped tokens, if the base token is in `ACCEPTABLE_HIGH_PRICES` we can exit the sense check

    elif bucket == 'yearn or yearn-like':
        try: # v2
            underlying = await raw_call_async(token_address, 'token()', output='address')
            if underlying in ACCEPTABLE_HIGH_PRICES or await _exit_sense_check(underlying):
                return True
        except:
            pass
        try: # v1
            underlying = await raw_call_async(token_address, 'want()', output='address')
            if underlying in ACCEPTABLE_HIGH_PRICES or await _exit_sense_check(underlying):
                return True
        except:
            pass
    
    elif bucket == 'atoken':
        try: # v2
            underlying = await raw_call_async(token_address, 'UNDERLYING_ASSET_ADDRESS()', output='address')
            if underlying in ACCEPTABLE_HIGH_PRICES or await _exit_sense_check(underlying):
                return True
        except:
            pass
        try: # v1
            underlying = await raw_call_async(token_address, 'underlyingAssetAddress()', output='address')
            if underlying in ACCEPTABLE_HIGH_PRICES or await _exit_sense_check(underlying):
                return True
        except:
            pass

    elif bucket == 'compound':
        underlying = await raw_call_async(token_address, 'underlying()', output='address')
        if underlying in ACCEPTABLE_HIGH_PRICES or await _exit_sense_check(underlying):
            return True
    
    return False




import logging
from typing import Optional

from async_lru import alru_cache
from y import convert
from y.contracts import Contract, has_methods_async
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils.logging import yLazyLogger
from y.utils.multicall import fetch_multicall

logger = logging.getLogger(__name__)

@yLazyLogger(logger)
@alru_cache(maxsize=None)
async def is_mstable_feeder_pool(address: AnyAddressType) -> bool:
    return await has_methods_async(address, ('getPrice()((uint,uint))','mAsset()(address)'))

@yLazyLogger(logger)
def get_price(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    contract = Contract(address)
    ratio, masset, decimals = fetch_multicall([contract,'getPrice'],[contract,'mAsset'],[contract,'decimals'],block=block)
    ratio = ratio[0] / 10 ** decimals
    underlying_price = magic.get_price(masset,block)
    return UsdPrice(underlying_price * ratio)



import pytest
from brownie import ZERO_ADDRESS, chain
from multicall import Call
from tests.test_constants import STABLECOINS
from tests.fixtures import blocks_for_contract, mutate_tokens
from y.classes.common import ERC20
from y.constants import WRAPPED_GAS_COIN, wbtc
from y.exceptions import NoProxyImplementation, call_reverted

TOKENS = list(STABLECOINS) + [WRAPPED_GAS_COIN, wbtc.address]

TOKENS_BY_BLOCK = [
    (token, block)
    for token in TOKENS
    for block in blocks_for_contract(token)
]

@pytest.mark.parametrize('token',TOKENS)
def test_erc20(token):
    token = ERC20(token)
    if token.address == '0x57ab1e02fee23774580c119740129eac7081e9d3': # old sUSD
        pytest.skip('Not applicable to deprecated sUSD.')
    block = chain.height
    assert token.contract, f'Cannot fetch contract for token {token}'
    assert token.build_name, f'Cannot fetch build name for token {token}'
    assert token.symbol, f'Cannot fetch symbol for token {token}'
    assert token.name, f'Cannot fetch name for token {token}'
    assert 10 ** token.decimals == token.scale, f'Incorrect scale fetched for token {token}'
    assert token.total_supply(block) / token.scale == token.total_supply_readable(block), f'Incorrect total supply readable for token {token}'
    assert token.price(), f'Cannot fetch price for token {token}'

@pytest.mark.parametrize('token,block',TOKENS_BY_BLOCK)
def test_erc20_at_block(token, block):
    token = ERC20(token)
    if token.address == '0x57ab1e02fee23774580c119740129eac7081e9d3' and block >= 13222927:
        pytest.skip('Not applicable to the old sUSD after migration block.')

    # NOTE Some proxy tokens would fail tests in early days because no implementation is specified.
    try:
        if Call(token.address, 'implementation()(address)', [['imp',None]], block_id = block)() == ZERO_ADDRESS:
            pytest.skip(f"Not applicable to proxy contracts with implementation not set.")
    except Exception as e:
        if not call_reverted(e):
            raise
    
    # NOTE We've validated token is not problematic proxy, proceed with test.
    try:
        # NOTE also tests ERC20._decimals
        assert token.total_supply(block) / token._scale(block) == token.total_supply_readable(block), f'Incorrect total supply readable for token {token}'
    except NoProxyImplementation:
        pass
    
    assert token.price(block), f'Cannot fetch price for token {token}'



import logging
from functools import cached_property, lru_cache
from typing import List, Optional

from async_lru import alru_cache
from brownie import chain
from brownie.convert.datatypes import EthAddress, HexString
from eth_abi import encode_single
from multicall import Call
from multicall.utils import await_awaitable
from y import convert
from y.classes.singleton import Singleton
from y.contracts import Contract, has_method_async
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.exceptions import UnsupportedNetwork
from y.networks import Network
from y.utils.multicall import fetch_multicall

logger = logging.getLogger(__name__)

addresses = {
    Network.Mainnet: '0x823bE81bbF96BEc0e25CA13170F5AaCb5B79ba83',
}


class Synthetix(metaclass=Singleton):
    def __init__(self) -> None:
        if chain.id not in addresses:
            raise UnsupportedNetwork("synthetix is not supported on this network")

    @lru_cache(maxsize=None)
    def get_address(self, name) -> Contract:
        """
        Get contract from Synthetix registry.
        See also https://docs.synthetix.io/addresses/
        """
        address_resolver = Contract(addresses[chain.id])
        address = address_resolver.getAddress(encode_single('bytes32', name.encode()))
        proxy = Contract(address)
        return Contract(proxy.target()) if hasattr(proxy, 'target') else proxy

    @cached_property
    def synths(self) -> List[EthAddress]:
        """
        Get target addresses of all synths.
        """
        proxy_erc20 = self.get_address('ProxyERC20')
        synths = fetch_multicall(
            *[
                [proxy_erc20, 'availableSynths', i]
                for i in range(proxy_erc20.availableSynthCount())
            ]
        )
        logger.info(f'loaded {len(synths)} synths')
        return synths
    '''
    def __contains__(self, token: AnyAddressType) -> bool:
        """
        Check if a token is a synth.
        """
        token = convert.to_address(token)
        if synthetix.get_currency_key(token):
            return True
        if has_method(token, 'target()(address)'):
            target = Call(token, 'target()(address)')()
            return target in synthetix.synths and Call(target, 'proxy()(address)')() == token
        return False
    '''
    async def is_synth(self, token: AnyAddressType) -> bool:
        """
        Check if a token is a synth.
        """
        token = convert.to_address(token)
        if await synthetix.get_currency_key_async(token):
            return True
        if await has_method_async(token, 'target()(address)'):
            target = await Call(token, 'target()(address)').coroutine()
            return target in synthetix.synths and await Call(target, 'proxy()(address)').coroutine() == token
        return False


    @lru_cache(maxsize=None)
    def get_currency_key(self, token: AnyAddressType) -> Optional[HexString]:
        return await_awaitable(self.get_currency_key_async(token))
    
    @alru_cache(maxsize=None)
    async def get_currency_key_async(self, token: AnyAddressType) -> Optional[HexString]:
        target = await Call(token, ['target()(address)']).coroutine() if await has_method_async(token,'target()(address)') else token
        return await Call(target, ['currencyKey()(bytes32)']).coroutine() if await has_method_async(token,'currencyKey()(bytes32)') else None

    def get_price(self, token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        return await_awaitable(self.get_price_async(token, block=block))
    
    async def get_price_async(self, token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        """
        Get a price of a synth in dollars.
        """
        token = convert.to_address(token)
        rates = self.get_address('ExchangeRates')
        key = await self.get_currency_key_async(token)
        try:
            return UsdPrice(await rates.rateForCurrency.coroutine(key, block_identifier=block) / 1e18)
        except ValueError:
            return None


try:
    synthetix = Synthetix()
except UnsupportedNetwork:
    synthetix = set()




from tests.fixtures import mainnet_only
from y.classes.common import ERC20
from y.constants import usdc
from y.prices.yearn import YearnInspiredVault


@mainnet_only
def test_share_price_yearn_v1():
    yusdc_v1 = '0xa2609B2b43AC0F5EbE27deB944d2a399C201E3dA'
    block = 14_500_000
    assert YearnInspiredVault(yusdc_v1).share_price(block) == 1.111974132096754

@mainnet_only
def test_underlying_yearn_v1():
    yusdc_v1 = '0xa2609B2b43AC0F5EbE27deB944d2a399C201E3dA'
    assert YearnInspiredVault(yusdc_v1).underlying == ERC20(usdc)

@mainnet_only
def test_share_price_yearn_v2():
    yvusdc_v2 = '0x5f18c75abdae578b483e5f43f12a39cf75b973a9'
    block = 14_500_000
    assert YearnInspiredVault(yvusdc_v2).share_price(block) == 1.096431

@mainnet_only
def test_underlying_yearn_v2():
    yvusdc_v2 = '0x5f18c75abdae578b483e5f43f12a39cf75b973a9'
    assert YearnInspiredVault(yvusdc_v2).underlying == ERC20(usdc)
